"""Tests for pactown registry."""

import tempfile
from pathlib import Path

from pactown.registry.models import Artifact, ArtifactVersion, RegistryStorage


def test_artifact_version_to_dict():
    version = ArtifactVersion(
        version="1.0.0",
        readme_content="# Test",
        checksum="abc123",
    )
    data = version.to_dict()
    assert data["version"] == "1.0.0"
    assert data["readme_content"] == "# Test"
    assert data["checksum"] == "abc123"


def test_artifact_version_from_dict():
    data = {
        "version": "2.0.0",
        "readme_content": "# Content",
        "checksum": "def456",
        "published_at": "2024-01-01T00:00:00",
        "metadata": {"key": "value"},
    }
    version = ArtifactVersion.from_dict(data)
    assert version.version == "2.0.0"
    assert version.metadata == {"key": "value"}


def test_artifact_full_name():
    artifact = Artifact(name="my-service", namespace="prod")
    assert artifact.full_name == "prod/my-service"


def test_artifact_add_version():
    artifact = Artifact(name="test", namespace="default")
    version = ArtifactVersion(
        version="1.0.0",
        readme_content="# Test",
        checksum="abc",
    )
    artifact.add_version(version)

    assert artifact.latest_version == "1.0.0"
    assert "1.0.0" in artifact.versions


def test_artifact_get_version():
    artifact = Artifact(name="test", namespace="default")
    v1 = ArtifactVersion(version="1.0.0", readme_content="v1", checksum="a")
    v2 = ArtifactVersion(version="2.0.0", readme_content="v2", checksum="b")

    artifact.add_version(v1)
    artifact.add_version(v2)

    assert artifact.get_version("1.0.0").readme_content == "v1"
    assert artifact.get_version("latest").readme_content == "v2"
    assert artifact.get_version("*").readme_content == "v2"
    assert artifact.get_version("3.0.0") is None


def test_registry_storage_save_and_get():
    with tempfile.TemporaryDirectory() as tmpdir:
        storage = RegistryStorage(Path(tmpdir))

        artifact = Artifact(name="test-svc", namespace="default")
        artifact.add_version(ArtifactVersion(
            version="1.0.0",
            readme_content="# Hello",
            checksum="abc",
        ))

        storage.save_artifact(artifact)

        # Retrieve
        retrieved = storage.get("default", "test-svc")
        assert retrieved is not None
        assert retrieved.name == "test-svc"
        assert retrieved.latest_version == "1.0.0"


def test_registry_storage_list():
    with tempfile.TemporaryDirectory() as tmpdir:
        storage = RegistryStorage(Path(tmpdir))

        for ns in ["default", "prod"]:
            artifact = Artifact(name=f"svc-{ns}", namespace=ns)
            storage.save_artifact(artifact)

        all_artifacts = storage.list()
        assert len(all_artifacts) == 2

        default_only = storage.list(namespace="default")
        assert len(default_only) == 1
        assert default_only[0].namespace == "default"


def test_registry_storage_delete():
    with tempfile.TemporaryDirectory() as tmpdir:
        storage = RegistryStorage(Path(tmpdir))

        artifact = Artifact(name="to-delete", namespace="default")
        storage.save_artifact(artifact)

        assert storage.get("default", "to-delete") is not None

        result = storage.delete("default", "to-delete")
        assert result is True
        assert storage.get("default", "to-delete") is None


def test_registry_storage_search():
    with tempfile.TemporaryDirectory() as tmpdir:
        storage = RegistryStorage(Path(tmpdir))

        a1 = Artifact(name="user-api", namespace="default", description="User management")
        a2 = Artifact(name="order-api", namespace="default", description="Order processing")
        a3 = Artifact(name="payment", namespace="default", tags=["api", "stripe"])

        storage.save_artifact(a1)
        storage.save_artifact(a2)
        storage.save_artifact(a3)

        results = storage.search("api")
        assert len(results) == 3  # matches name or tag

        results = storage.search("user")
        assert len(results) == 1
        assert results[0].name == "user-api"


def test_registry_storage_persistence():
    with tempfile.TemporaryDirectory() as tmpdir:
        # First instance
        storage1 = RegistryStorage(Path(tmpdir))
        artifact = Artifact(name="persistent", namespace="default")
        artifact.add_version(ArtifactVersion(
            version="1.0.0",
            readme_content="# Persist",
            checksum="xyz",
        ))
        storage1.save_artifact(artifact)

        # Second instance (simulating restart)
        storage2 = RegistryStorage(Path(tmpdir))
        retrieved = storage2.get("default", "persistent")

        assert retrieved is not None
        assert retrieved.latest_version == "1.0.0"



import json

import httpx
import pytest

from pactown.runner_api import RunnerApiSettings, RunnerService, create_runner_api
from pactown.service_runner import ErrorCategory, RunResult


def _sample_markdown() -> str:
    return """# Test Service

```python markpact:file path=main.py
print('hello')
```

```bash markpact:run
python main.py
```
"""


@pytest.mark.asyncio
async def test_validate_ok(tmp_path):
    settings = RunnerApiSettings()
    settings.require_token = False

    runner_service = RunnerService(sandbox_root=tmp_path, port_start=12000, port_end=12010)
    app = create_runner_api(runner_service=runner_service, settings=settings)

    transport = httpx.ASGITransport(app=app)
    async with httpx.AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.post("/validate", json={"readme_content": _sample_markdown()})
        assert resp.status_code == 200
        payload = resp.json()
        assert payload["valid"] is True
        assert payload["file_count"] == 1
        assert payload["has_run"] is True


@pytest.mark.asyncio
async def test_sandbox_prepare_and_file_ops(tmp_path):
    settings = RunnerApiSettings()
    settings.require_token = False

    runner_service = RunnerService(sandbox_root=tmp_path, port_start=12000, port_end=12010)
    app = create_runner_api(runner_service=runner_service, settings=settings)

    transport = httpx.ASGITransport(app=app)
    async with httpx.AsyncClient(transport=transport, base_url="http://test") as client:
        service_id = "1-user"

        prep = await client.post(
            "/sandbox/prepare",
            json={
                "project_id": 1,
                "service_id": service_id,
                "readme_content": _sample_markdown(),
                "port": 0,
            },
        )
        assert prep.status_code == 200
        prep_payload = prep.json()
        file_paths = [f["path"] for f in prep_payload.get("files", [])]
        assert "main.py" in file_paths

        read = await client.get(f"/sandbox/{service_id}/file", params={"path": "main.py"})
        assert read.status_code == 200
        assert "print('hello')" in read.json()["content"]

        write = await client.put(
            f"/sandbox/{service_id}/file",
            params={"path": "extra.txt"},
            json={"content": "ok"},
        )
        assert write.status_code == 200

        files = await client.get(f"/sandbox/{service_id}/files")
        assert files.status_code == 200
        file_paths = [f["path"] for f in files.json().get("files", [])]
        assert "extra.txt" in file_paths

        bad = await client.put(
            f"/sandbox/{service_id}/file",
            params={"path": "../escape.txt"},
            json={"content": "no"},
        )
        assert bad.status_code == 400

        delete = await client.delete(f"/sandbox/{service_id}/file", params={"path": "extra.txt"})
        assert delete.status_code == 200

        files2 = await client.get(f"/sandbox/{service_id}/files")
        assert files2.status_code == 200
        file_paths2 = [f["path"] for f in files2.json().get("files", [])]
        assert "extra.txt" not in file_paths2


@pytest.mark.asyncio
async def test_status_filtering(tmp_path):
    settings = RunnerApiSettings()
    settings.require_token = False

    runner_service = RunnerService(sandbox_root=tmp_path, port_start=12000, port_end=12010)
    app = create_runner_api(runner_service=runner_service, settings=settings)

    transport = httpx.ASGITransport(app=app)
    async with httpx.AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.get("/status", params={"user_id": "user:1"})
        assert resp.status_code == 200
        assert resp.json().get("services") == []


@pytest.mark.asyncio
async def test_run_failure_includes_error_report_md(tmp_path, monkeypatch):
    settings = RunnerApiSettings()
    settings.require_token = False

    runner_service = RunnerService(sandbox_root=tmp_path, port_start=12000, port_end=12010)
    app = create_runner_api(runner_service=runner_service, settings=settings)

    service_id = "1-user"
    sandbox_path = runner_service._sandbox_path_for(service_id)
    sandbox_path.mkdir(parents=True, exist_ok=True)
    (sandbox_path / "main.py").write_text("print('hello')\n", encoding="utf-8")

    file_path = sandbox_path / "main.py"
    stderr = (
        "Traceback (most recent call last):\n"
        f"  File \"{file_path}\", line 1, in <module>\n"
        "    raise RuntimeError('boom')\n"
        "RuntimeError: boom\n"
        "trace_id=abc123\n"
    )

    async def fake_run(
        *,
        service_id: str,
        content: str,
        port: int,
        env,
        user_id,
        username,
        user_profile,
        fast_mode: bool,
        skip_health_check: bool,
        on_log=None,
    ) -> RunResult:
        return RunResult(
            success=False,
            port=int(port or 0),
            pid=None,
            message="boom",
            logs=[],
            error_category=ErrorCategory.PROCESS_CRASH,
            stderr_output=stderr,
        )

    monkeypatch.setattr(runner_service, "run", fake_run)

    transport = httpx.ASGITransport(app=app)
    async with httpx.AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.post(
            "/run",
            json={
                "project_id": 1,
                "service_id": service_id,
                "readme_content": _sample_markdown(),
                "port": 12000,
            },
        )
        assert resp.status_code == 200
        payload = resp.json()
        assert payload.get("error_context") is not None
        assert payload.get("error_report_md")

        md = payload["error_report_md"]
        assert "## Summary" in md
        assert "abc123" in md
        assert "### `main.py`" in md
        assert "print('hello')" in md


@pytest.mark.asyncio
async def test_run_stream_failure_includes_error_report_md(tmp_path, monkeypatch):
    settings = RunnerApiSettings()
    settings.require_token = False

    runner_service = RunnerService(sandbox_root=tmp_path, port_start=12000, port_end=12010)
    app = create_runner_api(runner_service=runner_service, settings=settings)

    service_id = "1-user"
    sandbox_path = runner_service._sandbox_path_for(service_id)
    sandbox_path.mkdir(parents=True, exist_ok=True)
    (sandbox_path / "main.py").write_text("print('hello')\n", encoding="utf-8")

    file_path = sandbox_path / "main.py"
    stderr = (
        "Traceback (most recent call last):\n"
        f"  File \"{file_path}\", line 1, in <module>\n"
        "    raise RuntimeError('boom')\n"
        "RuntimeError: boom\n"
        "trace_id=abc123\n"
    )

    async def fake_run(
        *,
        service_id: str,
        content: str,
        port: int,
        env,
        user_id,
        username,
        user_profile,
        fast_mode: bool,
        skip_health_check: bool,
        on_log=None,
    ) -> RunResult:
        return RunResult(
            success=False,
            port=int(port or 0),
            pid=None,
            message="boom",
            logs=[],
            error_category=ErrorCategory.PROCESS_CRASH,
            stderr_output=stderr,
        )

    monkeypatch.setattr(runner_service, "run", fake_run)

    transport = httpx.ASGITransport(app=app)
    async with httpx.AsyncClient(transport=transport, base_url="http://test") as client:
        async with client.stream(
            "POST",
            "/run/stream",
            json={
                "project_id": 1,
                "service_id": service_id,
                "readme_content": _sample_markdown(),
                "port": 12000,
            },
        ) as resp:
            assert resp.status_code == 200
            items = []
            async for line in resp.aiter_lines():
                if not line:
                    continue
                items.append(json.loads(line))

        result_item = next(i for i in items if i.get("type") == "result")
        result_payload = result_item["result"]
        assert result_payload.get("error_report_md")
        assert "### `main.py`" in result_payload["error_report_md"]



import asyncio
import io
import os
from pathlib import Path
from types import SimpleNamespace

import pytest

from pactown.config import CacheConfig
from pactown.service_runner import ServiceRunner


def test_fast_run_fallback_sets_serviceconfig_readme_and_cleans_temp_file(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    runner = ServiceRunner(
        sandbox_root=tmp_path / "sandboxes",
        enable_fast_start=False,
        cache_config=CacheConfig(),
    )

    async def allow(*args, **kwargs):
        return SimpleNamespace(allowed=True, reason=None, delay_seconds=0.0)

    monkeypatch.setattr(runner.security_policy, "check_can_start_service", allow)

    from pactown import service_runner as sr_module

    monkeypatch.setattr(sr_module, "kill_process_on_port", lambda _port: False)

    captured = {}

    def fake_create_sandbox(config, readme_path, install_dependencies=True, on_log=None, env=None):
        captured["config_readme"] = config.readme
        captured["readme_path"] = readme_path
        captured["env"] = env
        # Readme file should exist while sandbox is being created
        assert Path(config.readme).exists()
        assert readme_path.exists()
        sandbox_path = tmp_path / "sandbox"
        sandbox_path.mkdir(parents=True, exist_ok=True)
        return SimpleNamespace(path=sandbox_path)

    monkeypatch.setattr(runner.sandbox_manager, "create_sandbox", fake_create_sandbox)
    runner.sandbox_manager._processes = {}

    class DummyPopen:
        def __init__(self):
            self.pid = 12345
            self.stdout = io.BytesIO(b"")
            self.stderr = io.BytesIO(b"")

        def poll(self):
            return None

    import subprocess as subprocess_module

    monkeypatch.setattr(subprocess_module, "Popen", lambda *a, **k: DummyPopen())

    content = """```python markpact:file path=main.py
print('hi')
```
```bash markpact:run
python main.py
```"""

    result = asyncio.run(
        runner.fast_run(
            service_id="svc",
            content=content,
            port=8123,
            env={"PIP_INDEX_URL": "http://pypi-proxy.local/simple"},
            skip_health_check=True,
        )
    )

    assert result.success is True
    assert "readme_path" in captured
    assert captured["config_readme"] == str(captured["readme_path"])
    assert captured["env"] == {"PIP_INDEX_URL": "http://pypi-proxy.local/simple"}

    # The temp file should be removed by the finally block
    assert not captured["readme_path"].exists()



from pactown.markpact_blocks import parse_blocks


def test_parse_blocks_new_format_includes_lang() -> None:
    md = """```python markpact:file path=main.py
print(\"hi\")
```"""

    blocks = parse_blocks(md)

    assert len(blocks) == 1

    block = blocks[0]
    assert block.kind == "file"
    assert block.meta == "path=main.py"
    assert block.lang == "python"
    assert block.body == 'print("hi")'
    assert block.get_path() == "main.py"


def test_parse_blocks_old_format_is_supported() -> None:
    md = """```markpact:file python path=main.py
print(\"hi\")
```"""

    blocks = parse_blocks(md)

    assert len(blocks) == 1

    block = blocks[0]
    assert block.kind == "file"
    assert block.meta == "python path=main.py"
    assert block.lang == ""
    assert block.body == 'print("hi")'
    assert block.get_path() == "main.py"


def test_parse_blocks_run_block_new_format() -> None:
    md = """```bash markpact:run
echo hi
```"""

    blocks = parse_blocks(md)

    assert len(blocks) == 1

    block = blocks[0]
    assert block.kind == "run"
    assert block.meta == ""
    assert block.lang == "bash"
    assert block.body == "echo hi"



"""
User isolation module for pactown.

Provides Linux user-based sandbox isolation for multi-tenant SaaS:
- Create isolated Linux users per SaaS user
- Run sandboxes under isolated user accounts
- Easy migration of projects between hosts
- Resource limits via cgroups (optional)
"""

import grp
import os
import pwd
import re
import shutil
import subprocess
import logging
import tempfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Any
from threading import Lock

logger = logging.getLogger("pactown.isolation")


def _sanitize_gecos(value: str) -> str:
    v = str(value or "")
    v = v.replace(":", "_")
    v = re.sub(r"[\x00\r\n\t]", " ", v)
    v = re.sub(r"\s+", " ", v).strip()
    if not v:
        v = "pactown-user"
    return v[:200]


@dataclass
class IsolatedUser:
    """Represents an isolated Linux user for sandbox execution."""
    saas_user_id: str
    linux_username: str
    linux_uid: int
    linux_gid: int
    home_dir: Path
    created_at: float = field(default_factory=lambda: __import__('time').time())
    
    def to_dict(self) -> dict:
        return {
            "saas_user_id": self.saas_user_id,
            "linux_username": self.linux_username,
            "linux_uid": self.linux_uid,
            "linux_gid": self.linux_gid,
            "home_dir": str(self.home_dir),
            "created_at": self.created_at,
        }


class UserIsolationManager:
    """
    Manages isolated Linux users for sandbox execution.
    
    Each SaaS user gets a dedicated Linux user account:
    - Username: pactown_<hash(saas_user_id)>
    - Home dir: /home/pactown_users/<username>
    - All sandboxes run under this user
    
    Benefits:
    - Process isolation (different UIDs)
    - File system isolation (home directories)
    - Easy migration (tar user's home dir)
    - Resource limits via cgroups
    """
    
    PREFIX = "pactown_"
    BASE_UID = 60000  # Start UIDs from 60000
    BASE_GID = 60000
    
    def __init__(
        self,
        users_base: Path = Path("/home/pactown_users"),
        enable_cgroups: bool = False,
    ):
        self.users_base = users_base
        self.enable_cgroups = enable_cgroups
        self._users: Dict[str, IsolatedUser] = {}
        self._lock = Lock()
        self._next_uid = self.BASE_UID
        
        # Create base directory if running as root
        if os.geteuid() == 0:
            self.users_base.mkdir(parents=True, exist_ok=True)
        
        # Load existing users
        self._load_existing_users()

    def can_isolate(self) -> tuple[bool, str]:
        if os.geteuid() != 0:
            return False, "not running as root"
        if shutil.which("useradd") is None:
            return False, "missing 'useradd' (install 'passwd'/'shadow' tools)"
        if shutil.which("groupadd") is None:
            return False, "missing 'groupadd' (install 'passwd'/'shadow' tools)"
        try:
            self.users_base.mkdir(parents=True, exist_ok=True)
        except Exception as e:
            return False, f"cannot create users_base={self.users_base}: {e}"
        try:
            if not os.access(self.users_base, os.W_OK | os.X_OK):
                return False, f"users_base not writable: {self.users_base}"
        except Exception as e:
            return False, f"cannot check permissions for users_base={self.users_base}: {e}"
        return True, ""
    
    def _load_existing_users(self):
        """Load existing pactown users from system."""
        try:
            for entry in pwd.getpwall():
                if entry.pw_name.startswith(self.PREFIX):
                    # Extract saas_user_id from comment field or username
                    saas_id = entry.pw_gecos or entry.pw_name[len(self.PREFIX):]
                    self._users[saas_id] = IsolatedUser(
                        saas_user_id=saas_id,
                        linux_username=entry.pw_name,
                        linux_uid=entry.pw_uid,
                        linux_gid=entry.pw_gid,
                        home_dir=Path(entry.pw_dir),
                    )
                    if entry.pw_uid >= self._next_uid:
                        self._next_uid = entry.pw_uid + 1
        except Exception as e:
            logger.warning(f"Could not load existing users: {e}")
    
    def _generate_username(self, saas_user_id: str) -> str:
        """Generate Linux username from SaaS user ID."""
        import hashlib
        hash_suffix = hashlib.sha256(saas_user_id.encode()).hexdigest()[:8]
        return f"{self.PREFIX}{hash_suffix}"
    
    def get_or_create_user(self, saas_user_id: str) -> IsolatedUser:
        """Get or create an isolated Linux user for a SaaS user."""
        with self._lock:
            if saas_user_id in self._users:
                return self._users[saas_user_id]
            
            # Create new user
            username = self._generate_username(saas_user_id)
            uid = self._next_uid
            gid = self._next_uid
            home_dir = self.users_base / username

            # If user already exists (deterministic username), reuse it.
            try:
                existing = pwd.getpwnam(username)
                user = IsolatedUser(
                    saas_user_id=saas_user_id,
                    linux_username=existing.pw_name,
                    linux_uid=existing.pw_uid,
                    linux_gid=existing.pw_gid,
                    home_dir=Path(existing.pw_dir),
                )
                self._users[saas_user_id] = user
                if existing.pw_uid >= self._next_uid:
                    self._next_uid = existing.pw_uid + 1
                logger.info(
                    "Reusing existing Linux user %s (uid=%s) for %s",
                    user.linux_username,
                    user.linux_uid,
                    saas_user_id,
                )
                return user
            except KeyError:
                pass
            
            # Check if we can create users (requires root)
            if os.geteuid() != 0:
                # Non-root mode: create virtual user for tracking
                logger.warning(f"Not running as root, creating virtual user for {saas_user_id}")
                user = IsolatedUser(
                    saas_user_id=saas_user_id,
                    linux_username=username,
                    linux_uid=os.getuid(),  # Use current user
                    linux_gid=os.getgid(),
                    home_dir=home_dir,
                )
                self._users[saas_user_id] = user
                
                # Create home directory anyway
                try:
                    home_dir.mkdir(parents=True, exist_ok=True)
                except Exception as e:
                    fallback_base = Path(tempfile.gettempdir()) / "pactown_users"
                    fallback_home = fallback_base / username
                    logger.warning(
                        "Could not create users_base home_dir=%s (%s); falling back to %s",
                        home_dir,
                        e,
                        fallback_home,
                    )
                    fallback_home.mkdir(parents=True, exist_ok=True)
                    user.home_dir = fallback_home
                return user

            can_isolate, reason = self.can_isolate()
            if not can_isolate:
                raise RuntimeError(f"User isolation unavailable: {reason}")
            
            try:
                # Create group
                try:
                    grp.getgrnam(username)
                except KeyError:
                    res = subprocess.run(
                        ["groupadd", "-g", str(gid), username],
                        check=True,
                        capture_output=True,
                        text=True,
                    )
                    if res.stdout:
                        logger.debug("groupadd stdout: %s", res.stdout.strip())
                    if res.stderr:
                        logger.debug("groupadd stderr: %s", res.stderr.strip())
                
                # Create user
                safe_comment = _sanitize_gecos(saas_user_id)
                res = subprocess.run(
                    [
                        "useradd",
                        "-u",
                        str(uid),
                        "-g",
                        str(gid),
                        "-d",
                        str(home_dir),
                        "-m",
                        "-s",
                        "/bin/bash",
                        "-c",
                        safe_comment,
                        username,
                    ],
                    check=True,
                    capture_output=True,
                    text=True,
                )
                if res.stdout:
                    logger.debug("useradd stdout: %s", res.stdout.strip())
                if res.stderr:
                    logger.debug("useradd stderr: %s", res.stderr.strip())
                
                logger.info(f"Created Linux user {username} (uid={uid}) for {saas_user_id}")
                
            except subprocess.CalledProcessError as e:
                stderr = getattr(e, "stderr", None)
                stdout = getattr(e, "stdout", None)
                logger.error(
                    "Failed to create isolated user for %s: cmd=%s returncode=%s stdout=%s stderr=%s",
                    saas_user_id,
                    getattr(e, "cmd", None),
                    getattr(e, "returncode", None),
                    (stdout.strip() if isinstance(stdout, str) else stdout),
                    (stderr.strip() if isinstance(stderr, str) else stderr),
                )
                raise RuntimeError(
                    f"Failed to create isolated user (saas_user_id={saas_user_id}, username={username}): {e}"
                )
            
            user = IsolatedUser(
                saas_user_id=saas_user_id,
                linux_username=username,
                linux_uid=uid,
                linux_gid=gid,
                home_dir=home_dir,
            )
            
            self._users[saas_user_id] = user
            self._next_uid += 1
            
            return user
    
    def get_user(self, saas_user_id: str) -> Optional[IsolatedUser]:
        """Get isolated user without creating."""
        return self._users.get(saas_user_id)
    
    def get_sandbox_path(self, saas_user_id: str, service_id: str) -> Path:
        """Get sandbox path for a specific service under user's home."""
        user = self.get_or_create_user(saas_user_id)
        sandbox_dir = user.home_dir / "sandboxes" / service_id
        sandbox_dir.mkdir(parents=True, exist_ok=True)
        
        # Set ownership if root
        if os.geteuid() == 0:
            os.chown(sandbox_dir, user.linux_uid, user.linux_gid)
        
        return sandbox_dir
    
    def run_as_user(
        self,
        saas_user_id: str,
        command: str,
        cwd: Path,
        env: Optional[Dict[str, str]] = None,
    ) -> subprocess.Popen:
        """
        Run a command as the isolated user.
        
        Returns subprocess.Popen for the running process.
        """
        user = self.get_or_create_user(saas_user_id)
        
        full_env = os.environ.copy()
        full_env["HOME"] = str(user.home_dir)
        full_env["USER"] = user.linux_username
        full_env["LOGNAME"] = user.linux_username
        if env:
            full_env.update(env)
        
        def set_user():
            """Pre-exec function to switch user."""
            if os.geteuid() == 0:
                os.setgid(user.linux_gid)
                os.setuid(user.linux_uid)
        
        # nosec B602: shell=True required for user commands in isolated sandbox
        # User is isolated via Linux user/group and sandbox directory
        process = subprocess.Popen(
            command,
            shell=True,  # nosec B602
            cwd=str(cwd),
            env=full_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=set_user if os.geteuid() == 0 else None,
        )
        
        logger.info(f"Started process {process.pid} as user {user.linux_username}")
        return process
    
    def list_users(self) -> List[IsolatedUser]:
        """List all isolated users."""
        return list(self._users.values())
    
    def get_user_stats(self, saas_user_id: str) -> Dict[str, Any]:
        """Get stats for a user's sandboxes."""
        user = self.get_user(saas_user_id)
        if not user:
            return {"error": "User not found"}
        
        sandboxes_dir = user.home_dir / "sandboxes"
        if not sandboxes_dir.exists():
            return {
                "user": user.to_dict(),
                "sandbox_count": 0,
                "total_size_mb": 0,
            }
        
        sandbox_count = len(list(sandboxes_dir.iterdir()))
        total_size = sum(
            f.stat().st_size 
            for f in sandboxes_dir.rglob("*") 
            if f.is_file()
        )
        
        return {
            "user": user.to_dict(),
            "sandbox_count": sandbox_count,
            "total_size_mb": total_size / (1024 * 1024),
        }
    
    def export_user_data(self, saas_user_id: str, output_path: Path) -> bool:
        """
        Export user's data for migration.
        
        Creates a tar.gz archive of the user's home directory.
        """
        user = self.get_user(saas_user_id)
        if not user or not user.home_dir.exists():
            return False
        
        try:
            subprocess.run(
                ["tar", "-czf", str(output_path), "-C", str(user.home_dir.parent), user.linux_username],
                check=True,
                capture_output=True,
            )
            logger.info(f"Exported user {saas_user_id} to {output_path}")
            return True
        except subprocess.CalledProcessError as e:
            logger.error(f"Failed to export user: {e}")
            return False
    
    def import_user_data(self, saas_user_id: str, archive_path: Path) -> bool:
        """
        Import user's data from migration archive.
        """
        user = self.get_or_create_user(saas_user_id)
        
        try:
            # Extract to user's home
            subprocess.run(
                ["tar", "-xzf", str(archive_path), "-C", str(self.users_base)],
                check=True,
                capture_output=True,
            )
            
            # Fix ownership
            if os.geteuid() == 0:
                subprocess.run(
                    ["chown", "-R", f"{user.linux_uid}:{user.linux_gid}", str(user.home_dir)],
                    check=True,
                    capture_output=True,
                )
            
            logger.info(f"Imported user {saas_user_id} from {archive_path}")
            return True
        except subprocess.CalledProcessError as e:
            logger.error(f"Failed to import user: {e}")
            return False
    
    def delete_user(self, saas_user_id: str, delete_home: bool = True) -> bool:
        """Delete an isolated user."""
        user = self.get_user(saas_user_id)
        if not user:
            return False
        
        try:
            if os.geteuid() == 0:
                # Delete Linux user
                cmd = ["userdel"]
                if delete_home:
                    cmd.append("-r")
                cmd.append(user.linux_username)
                subprocess.run(cmd, check=True, capture_output=True)
                subprocess.run(
                    ["groupdel", user.linux_username],
                    capture_output=True,  # May fail if group doesn't exist
                )
            elif delete_home and user.home_dir.exists():
                # Non-root: just delete home directory
                shutil.rmtree(user.home_dir)
            
            del self._users[saas_user_id]
            logger.info(f"Deleted user {saas_user_id}")
            return True
            
        except subprocess.CalledProcessError as e:
            logger.error(f"Failed to delete user: {e}")
            return False


# Global isolation manager instance
_isolation_manager: Optional[UserIsolationManager] = None


def get_isolation_manager() -> UserIsolationManager:
    """Get global isolation manager instance."""
    global _isolation_manager
    if _isolation_manager is None:
        default_base = "/home/pactown_users" if os.geteuid() == 0 else tempfile.gettempdir() + "/pactown_users"
        users_base = Path(os.environ.get("PACTOWN_USERS_BASE", default_base))
        _isolation_manager = UserIsolationManager(users_base=users_base)
    return _isolation_manager



import pytest
from pactown.service_runner import ServiceRunner, ValidationResult

def test_validate_content_dependency_mismatch():
    runner = ServiceRunner(enable_fast_start=False)

    # Case 1: Node package in Python block
    content_node_in_py = """
```python markpact:file path=main.py
print("hello")
```
```python markpact:deps
express
requests
```
```bash markpact:run
python main.py
```
"""
    result = runner.validate_content(content_node_in_py)
    # Should be valid but contain warnings in errors list if strict validation is off, 
    # OR invalid if warnings are treated as errors. 
    # In my implementation: valid=len(errors) == 0 or all(e.startswith("Warning:") ...)
    # So valid should be True, but errors should contain the warning.
    assert result.valid is True
    assert any("Found Node.js package 'express' in Python dependency block" in e for e in result.errors)

    # Case 2: Python package in Node block
    content_py_in_node = """
```javascript markpact:file path=index.js
console.log("hello")
```
```javascript markpact:deps
fastapi
axios
```
```bash markpact:run
node index.js
```
"""
    result = runner.validate_content(content_py_in_node)
    assert result.valid is True
    assert any("Found Python package 'fastapi' in Node.js dependency block" in e for e in result.errors)

    # Case 3: Valid Python
    content_valid_py = """
```python markpact:file path=main.py
print("hello")
```
```python markpact:deps
fastapi
uvicorn
```
```bash markpact:run
python main.py
```
"""
    result = runner.validate_content(content_valid_py)
    assert result.valid is True
    assert len(result.errors) == 0

    # Case 4: Valid Node
    content_valid_node = """
```javascript markpact:file path=index.js
console.log("hello")
```
```javascript markpact:deps
express
pg
```
```bash markpact:run
node index.js
```
"""
    result = runner.validate_content(content_valid_node)
    assert result.valid is True
    assert len(result.errors) == 0
